@@ -96,10 +96,9 @@ class Kazure(object):
                disks=[{'size': 10}], disksize=10, diskthin=True, diskinterface='virtio', nets=['default'], iso=None,
                vnc=True, cloudinit=True, reserveip=False, reservedns=False, reservehost=False, start=True, keys=[],
                cmds=[], ips=None, netmasks=None, gateway=None, nested=True, dns=None, domain=None, tunnel=False,
-               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[], kernel=None,
-               initrd=None, cmdline=None, placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False,
-               numamode=None, numa=[], pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[],
-               vmuser=None):
+               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[], cmdline=None,
+               placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False, numamode=None, numa=[],
+               pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[], vmuser=None, guestagent=True):
         if self.exists(name):
             return {'result': 'failure', 'reason': f"VM {name} already exists"}
         if flavor is None:
@@ -732,6 +731,21 @@ class Kazure(object):
             disks[disk.name] = {'pool': 'default', 'path': disk.name}
         return disks
 
+    def detach_disks(self, name):
+        try:
+            vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
+        except:
+            error(f"VM {name} not found")
+            return {'result': 'failure', 'reason': f"VM {name} not found"}
+        if os.path.basename(vm.instance_view.statuses[1].code) == 'running':
+            result = self.compute_client.virtual_machines.begin_deallocate(self.resource_group, name)
+            result.wait()
+        if len(vm.storage_profile.data_disks) > 1:
+            vm.storage_profile.data_disks = vm.storage_profile.data_disks[:1]
+            result = self.compute_client.virtual_machines.begin_create_or_update(self.resource_group, name, vm)
+            result.wait()
+        return {'result': 'success'}
+
     def add_nic(self, name, network, model='virtio'):
         try:
             vm = self.compute_client.virtual_machines.get(self.resource_group, name, expand='instanceView')
@@ -789,7 +803,7 @@ class Kazure(object):
         else:
             return {'result': 'failure', 'reason': f'Image {image} not found'}
 
-    def add_image(self, url, pool, short=None, cmd=None, name=None, size=None, convert=False):
+    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
         bucket = None
         # if 'blob.core.windows.net' not in url:
         if self.storage_account not in url:
