@@ -3,18 +3,20 @@ from binascii import hexlify
 from cryptography import x509
 from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import hashes
-from ipaddress import ip_address, ip_network
+from getpass import getpass
 import json
 from kvirt import common
-from kvirt.common import error, pprint, warning
+from kvirt.common import error, pprint, warning, sdn_ip
 from kvirt.defaults import UBUNTUS, METADATA_FIELDS
 from kvirt.providers.vsphere.helpers import find, collectproperties, findvm, createfolder, changecd, convert, waitForMe
 from kvirt.providers.vsphere.helpers import createscsispec, creatediskspec, createdvsnicspec, createclonespec
-from kvirt.providers.vsphere.helpers import createnicspec, createisospec, deletedirectory, dssize, keep_lease_alive
+from kvirt.providers.vsphere.helpers import createnicspec, createisospec, createcdspec, deletedirectory
+from kvirt.providers.vsphere.helpers import dssize, keep_lease_alive, folder_exists
 from kvirt.providers.vsphere.helpers import create_filter_spec, get_all_obj, convert_properties, findvm2, findvmdc
 import os
 from pyVmomi import vim, vmodl
 from pyVim import connect
+from pyVim.task import WaitForTask
 import random
 import re
 import ssl
@@ -32,14 +34,11 @@ import webbrowser
 from zipfile import ZipFile
 
 
-def sdn_ip(ip, kubetype):
-    return kubetype is not None and kubetype == 'openshift' and ip_address(ip) in ip_network('10.132.0.0/14')
-
-
 class Ksphere:
     def __init__(self, host, user, password, datacenter, cluster, debug=False, isofolder=None,
                  filtervms=False, filteruser=False, filtertag=None, category='kcli', basefolder=None, dvs=True,
-                 import_network='VM Network', timeout=2700, force_pool=False, restricted=False, serial=False):
+                 import_network='VM Network', timeout=3600, force_pool=False, restricted=False, serial=False):
+        password = password or getpass()
         if timeout < 1:
             smart_stub = connect.SmartStubAdapter(host=host, port=443, sslContext=_create_unverified_context(),
                                                   connectionPoolTimeout=0)
@@ -73,7 +72,8 @@ class Ksphere:
         self.networks = []
         self.dvs = dvs
         self.portgs = {}
-        self.restricted = restricted
+        self.esx = '.' in cluster and datacenter == 'ha-datacenter'
+        self.restricted = restricted or self.esx
         self.serial = serial
         self.import_network = import_network
         self.force_pool = force_pool
@@ -129,8 +129,7 @@ class Ksphere:
         si = self.si
         dc = self.dc
         vmFolder = dc.vmFolder
-        vm = findvm(si, vmFolder, name)
-        return True if vm is not None else False
+        return findvm(si, vmFolder, name) is not None
 
     def net_exists(self, name):
         print("not implemented")
@@ -140,10 +139,12 @@ class Ksphere:
                disks=[{'size': 10}], disksize=10, diskthin=True, diskinterface='virtio', nets=['default'], iso=None,
                vnc=True, cloudinit=True, reserveip=False, reservedns=False, reservehost=False, start=True, keys=[],
                cmds=[], ips=None, netmasks=None, gateway=None, nested=True, dns=None, domain=None, tunnel=False,
-               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[], kernel=None,
-               initrd=None, cmdline=None, placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False,
+               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[],
+               cmdline=None, placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False,
                numamode=None, numa=[], pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[],
-               vmuser=None):
+               vmuser=None, guestagent=True):
+        if not self.esx and self.exists(name):
+            return {'result': 'failure', 'reason': f"VM {name} already exists"}
         dc = self.dc
         vmFolder = dc.vmFolder
         diskmode = 'persistent'
@@ -181,18 +182,21 @@ class Ksphere:
             image = os.path.basename(image)
             clonespec = createclonespec(resourcepool)
             rootFolder = self.rootFolder
-            imageobj, imagedc = findvmdc(si, rootFolder, image, self.dc)
-            if imageobj is None:
+            imageobj, imagedc = findvmdc(si, rootFolder, image, dc)
+            if self.esx:
+                imagepool = pool
+                imagedc = dc.name
+            elif imageobj is None:
                 return {'result': 'failure', 'reason': f"Image {image} not found"}
             datastores = self._datastores_datacenters()
-            if os.path.basename(datastores[pool]) != self.dc.name:
-                return {'result': 'failure', 'reason': f"Pool {pool} doesn't belong to Datacenter {self.dc.name}"}
+            if os.path.basename(datastores[pool]) != dc.name:
+                return {'result': 'failure', 'reason': f"Pool {pool} doesn't belong to Datacenter {dc.name}"}
             if imagepool is None:
                 devices = imageobj.config.hardware.device
                 for number, dev in enumerate(devices):
                     if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
                         imagepool = dev.backing.datastore.name
-            if imagedc != self.dc.name or (overrides.get('force_pool', self.force_pool) and imagepool != pool):
+            if imagedc != dc.name or (overrides.get('force_pool', self.force_pool) and imagepool != pool):
                 warning(f"Vm {name} will be relocated from pool {imagepool} to {pool}")
                 relospec = vim.vm.RelocateSpec()
                 relospec.datastore = find(si, rootFolder, vim.Datastore, pool)
@@ -202,8 +206,19 @@ class Ksphere:
             confspec.flags = vim.vm.FlagInfo()
             confspec.flags.diskUuidEnabled = True
             confspec.annotation = name
+            if memory % 1024 != 0:
+                warning("Rounding up memory to be multiple of 1024")
+                memory += (1024 - memory) % 1024
             confspec.memoryMB = memory
             confspec.memoryHotAddEnabled = memoryhotplug
+            cores = overrides.get('cores')
+            sockets, cores = overrides.get('sockets', 1), overrides.get('cores')
+            if cores is not None and isinstance(cores, int) and isinstance(sockets, int):
+                confspec.numCoresPerSocket = cores
+                numcpus = cores * sockets
+            threads = overrides.get('threads')
+            if threads is not None and isinstance(threads, int):
+                confspec.simultaneousThreads = threads
             confspec.numCPUs = numcpus
             confspec.cpuHotAddEnabled = cpuhotplug
             confspec.cpuHotRemoveEnabled = cpuhotplug
@@ -270,9 +285,13 @@ class Ksphere:
                     guestopt.value = overrides[key]
                     extraconfig.append(guestopt)
             confspec.extraConfig = extraconfig
-            t = imageobj.CloneVM_Task(folder=vmfolder, name=name, spec=clonespec)
+            if not self.esx:
+                t = imageobj.CloneVM_Task(folder=vmfolder, name=name, spec=clonespec)
+            else:
+                vm = findvm(si, vmFolder, name)
+                t = vm.ReconfigVM_Task(confspec)
             waitForMe(t)
-            if cloudinitiso is not None:
+            if cloudinit and cloudinitiso is not None:
                 with TemporaryDirectory() as tmpdir:
                     if combustion:
                         cmdsdata = common.process_combustion_cmds(cmds, overrides)
@@ -281,13 +300,12 @@ class Ksphere:
                                 combustionfile.write(cmdsdata)
                     common.make_iso(name, tmpdir, userdata, meta, netdata, openstack=False, combustion=combustion)
                     cloudinitisofile = f"{tmpdir}/{name}.ISO"
+                    isopool = default_pool
+                    isofolder = None
                     if self.isofolder is not None:
                         isofolder = self.isofolder.split('/')
                         isopool = re.sub(r"[\[\]]", '', isofolder[0])
                         isofolder = isofolder[1]
-                    else:
-                        isopool = default_pool
-                        isofolder = None
                     self._uploadimage(isopool, cloudinitisofile, name, isofolder=isofolder)
                 vm = findvm(si, vmFolder, name)
                 c = changecd(self.si, vm, cloudinitiso)
@@ -338,6 +356,7 @@ class Ksphere:
         devconfspec = []
         unit_number = 1
         for index, disk in enumerate(disks):
+            diskuuid = None
             if disk is None:
                 disksize = default_disksize
                 diskthin = default_diskthin
@@ -358,6 +377,13 @@ class Ksphere:
                 diskthin = disk.get('thin', default_diskthin)
                 diskinterface = disk.get('interface', default_diskinterface)
                 diskpool = disk.get('pool', default_pool)
+                diskuuid = disk.get('uuid') or disk.get('wwn')
+                if diskuuid is not None:
+                    try:
+                        UUID(diskuuid)
+                    except:
+                        warning(f"{diskuuid} it not a valid disk uuid")
+                        diskuuid = None
             if index < len(currentdisks) and image is not None:
                 currentdisk = currentdisks[index]
                 currentsize = convert(1000 * currentdisk.capacityInKB, GB=False)
@@ -375,19 +401,45 @@ class Ksphere:
                     return {'result': 'failure', 'reason': f"Pool {diskpool} not found"}
                 else:
                     datastores[diskpool] = datastore
+            kept_disks = folder_exists(datastores[diskpool], f'{name}_keep')
             if index == 0:
                 scsispec = createscsispec()
                 devconfspec.append(scsispec)
+            elif kept_disks:
+                pprint("Using existing data disks")
+                path = f"[{diskpool}] {name}"
+                keep_path = f"[{diskpool}] {name}_keep"
+                browser = datastore.browser
+                task = browser.SearchDatastore_Task(datastorePath=keep_path, searchSpec=None)
+                fileManager = self.si.content.fileManager
+                WaitForTask(task)
+                result = task.info.result
+                files = result.file if hasattr(result, 'file') else []
+                for file in files:
+                    t = fileManager.MoveDatastoreFile_Task(sourceName=f"{keep_path}/{file.path}",
+                                                           sourceDatacenter=dc,
+                                                           destinationName=f"{path}/{file.path}",
+                                                           destinationDatacenter=dc,
+                                                           force=True)
+                    WaitForTask(t)
+                task = fileManager.DeleteDatastoreFile_Task(name=keep_path, datacenter=dc)
+                WaitForTask(task)
             if unit_number == 7:
                 unit_number = 8
-            diskspec = creatediskspec(unit_number, disksize, datastore, diskmode, diskthin)
+            diskpath = f'{name}/{name}_{index}.vmdk' if kept_disks else None
+            diskspec = creatediskspec(unit_number, disksize, datastore, diskmode, diskthin, diskuuid, diskpath)
             devconfspec.append(diskspec)
             unit_number += 1
+        for index, _ in enumerate(disks):
+            confspec.extraConfig.append(vim.option.OptionValue(key=f'scsi0:{index}.enableUUID', value='TRUE'))
         # NICSPEC
         if not self.networks:
             self.set_networks()
+        macs = {}
         for index, net in enumerate(nets):
             netname = net['name'] if isinstance(net, dict) else net
+            if isinstance(net, dict) and 'mac' in net:
+                macs[index] = net['mac']
             if netname == 'default':
                 if image is not None:
                     continue
@@ -414,16 +466,14 @@ class Ksphere:
                         currentnic.backing.port.switchUuid = switchuuid
                         currentnic.backing.port.portgroupKey = portgroupkey
                         currentnic.backing.port.portKey = None
-                        nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
-                        devconfspec.append(nicspec)
                     elif netname in self.networks:
                         currentnic.backing.deviceName = netname
-                        nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
-                        devconfspec.append(nicspec)
                     else:
                         return {'result': 'failure', 'reason': f"Invalid network {netname}"}
+                    nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
+                    devconfspec.append(nicspec)
                 continue
-            nicname = 'Network Adapter %d' % (index + 1)
+            nicname = f'Network Adapter {index + 1}'
             nictype = net['type'] if isinstance(net, dict) and 'type' in net else None
             if netname in self.portgs:
                 switchuuid = self.portgs[netname][0]
@@ -442,7 +492,7 @@ class Ksphere:
             else:
                 return {'result': 'failure', 'reason': f"Iso {iso} not found"}
         if need_cdrom:
-            cdspec = createisospec(iso)
+            cdspec = createcdspec() if iso is None and self.esx else createisospec(iso)
             devconfspec.append(cdspec)
         serial = overrides.get('serial', self.serial)
         if serial:
@@ -457,8 +507,25 @@ class Ksphere:
         confspec.deviceChange = devconfspec
         if nested:
             confspec.nestedHVEnabled = True
+        uefi = overrides.get('uefi', False)
+        uefi_legacy = overrides.get('uefi_legacy', False)
+        secureboot = overrides.get('secureboot', False)
+        if secureboot or uefi or uefi_legacy:
+            confspec.firmware = 'efi'
+            if secureboot:
+                confspec.bootOptions = vim.vm.BootOptions(efiSecureBootEnabled=True)
+        confspec.vAppConfigRemoved = True
         t = vm.Reconfigure(confspec)
         waitForMe(t)
+        if overrides.get('boot_order', False):
+            key_disks = [d.key for d in vm.config.hardware.device if isinstance(d, vim.vm.device.VirtualDisk)]
+            boot_disks = [vim.vm.BootOptions.BootableDiskDevice(deviceKey=key) for key in key_disks]
+            confspec = vim.vm.ConfigSpec()
+            confspec.bootOptions = vim.vm.BootOptions(bootOrder=boot_disks)
+            t = vm.Reconfigure(confspec)
+            waitForMe(t)
+        if macs:
+            self.set_macs(name, macs)
         if 'vmgroup' in overrides:
             vmgroup = overrides['vmgroup']
             vmgroups = {}
@@ -713,6 +780,21 @@ class Ksphere:
                 print(consolecommand)
             if not os.path.exists("/i_am_a_container"):
                 os.popen(consolecommand)
+        elif self.esx:
+            vmnumber = None
+            devices = config.hardware.device
+            for dev in devices:
+                if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
+                    vmnumber = dev.diskObjectId.split('-')[0]
+                    break
+            if vmnumber is not None:
+                vmurl = f"https://{self.vcip}/ui/#/console/{vmnumber}"
+                if self.debug or os.path.exists("/i_am_a_container"):
+                    msg = f"Open the following url:\n{vmurl}" if os.path.exists("/i_am_a_container") else vmurl
+                    pprint(msg)
+                else:
+                    pprint(f"Opening url {vmurl}")
+                    webbrowser.open(vmurl, new=2, autoraise=True)
         else:
             content = si.RetrieveContent()
             sgid = content.about.instanceUuid
@@ -773,14 +855,18 @@ class Ksphere:
                 yamlinfo['user'] = common.get_user(entry.value)
             if entry.key == 'tags':
                 yamlinfo['tags'] = entry.value
+            if entry.key == 'ip':
+                yamlinfo['ip'] = entry.value
         kubetype = yamlinfo.get('kubetype')
+        cluster_network = yamlinfo.get('cluster_network')
         ips = []
         if runtime.powerState == "poweredOn":
             yamlinfo['host'] = runtime.host.name
             for nic in guest.net:
                 if nic.ipAddress:
                     ip = nic.ipAddress[0]
-                    if not ip.startswith('fe80::') and not ip.startswith('169.254') and not sdn_ip(ip, kubetype):
+                    if not ip.startswith('fe80::') and not ip.startswith('169.254') and not sdn_ip(ip, kubetype,
+                                                                                                   cluster_network):
                         ips.append(ip)
                         if 'ip' not in yamlinfo:
                             yamlinfo['ip'] = ip
@@ -861,7 +947,6 @@ class Ksphere:
         pools = []
         rootFolder = self.rootFolder
         si = self.si
-        # dc = self.dc
         clu = find(si, rootFolder, vim.ComputeResource, self.clu)
         for dts in clu.datastore:
             pools.append(dts.name)
@@ -909,18 +994,15 @@ class Ksphere:
             waitForMe(t)
             result = t.info.result
             for element in result.file:
-                # if element.path.endswith('.iso'):
-                #    isos.append(f"{datastorepath}/{element.path}/{element.path}")
                 folderpath = element.path
-                if 'iso' not in folderpath.lower():
-                    continue
-                t = browser.SearchDatastoreSubFolders_Task(f"{datastorepath}{folderpath}", searchspec)
-                waitForMe(t)
-                for r in t.info.result:
-                    for isofile in r.file:
-                        iso_path = isofile.path
-                        if iso_path.endswith('.iso'):
-                            isos.append(f"{datastorepath}/{folderpath}/{iso_path}")
+                if 'iso' in folderpath.lower():
+                    t = browser.SearchDatastoreSubFolders_Task(f"{datastorepath}{folderpath}", searchspec)
+                    waitForMe(t)
+                    for r in t.info.result:
+                        for isofile in r.file:
+                            iso_path = f"{datastorepath}/{folderpath}/{isofile.path}"
+                            if iso_path.endswith('.iso') and iso_path not in isos:
+                                isos.append(iso_path)
         return isos
 
     def volumes(self, iso=False):
@@ -948,7 +1030,9 @@ class Ksphere:
         vmFolder = dc.vmFolder
         vm = findvm(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         configspec = vim.vm.ConfigSpec()
         opt = vim.option.OptionValue()
         opt.key = metatype
@@ -959,13 +1043,16 @@ class Ksphere:
 
     def update_memory(self, name, memory):
         if memory % 1024 != 0:
-            return {'result': 'failure', 'reason': "Memory needs to be multiple of 1024"}
+            warning("Rounding up memory to be multiple of 1024")
+            memory += (1024 - memory) % 1024
         si = self.si
         dc = self.dc
         vmFolder = dc.vmFolder
         vm = findvm(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         configspec = vim.vm.ConfigSpec()
         configspec.memoryMB = memory
         t = vm.ReconfigVM_Task(configspec)
@@ -977,7 +1064,9 @@ class Ksphere:
         vmFolder = dc.vmFolder
         vm = findvm(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         configspec = vim.vm.ConfigSpec()
         configspec.numCPUs = numcpus
         t = vm.ReconfigVM_Task(configspec)
@@ -1003,7 +1092,9 @@ class Ksphere:
             else:
                 iso = isos[0]
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         c = changecd(self.si, vm, iso)
         if iso is None:
             answered = False
@@ -1023,13 +1114,19 @@ class Ksphere:
         return {'result': 'success'}
 
     def convert_to_template(self, name):
+        if self.esx:
+            msg = "Operation not supported on single ESX"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         si = self.si
         dc = self.dc
         vmFolder = dc.vmFolder
         vm = findvm(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
-        else:
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
+        elif not self.esx:
             vm.MarkAsTemplate()
 
     def convert_to_vm(self, name):
@@ -1038,7 +1135,9 @@ class Ksphere:
         vmFolder = dc.vmFolder
         vm = findvm(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         else:
             rootFolder = self.rootFolder
             clu = find(si, rootFolder, vim.ComputeResource, self.clu)
@@ -1049,6 +1148,7 @@ class Ksphere:
 
     def _uploadimage(self, pool, origin, directory, isofolder=None):
         si = self.si
+        dc = self.dc
         rootFolder = self.rootFolder
         datastore = find(si, rootFolder, vim.Datastore, pool)
         if not datastore:
@@ -1058,7 +1158,7 @@ class Ksphere:
         destination = os.path.basename(origin)
         if isofolder is not None:
             directory = isofolder
-        url = f"https://{self.vcip}:443/folder/{directory}/{destination}?dcPath={self.dc.name}&dsName={pool}"
+        url = f"https://{self.vcip}:443/folder/{directory}/{destination}?dcPath={dc.name}&dsName={pool}"
         client_cookie = si._stub.cookie
         cookie_name = client_cookie.split("=", 1)[0]
         cookie_value = client_cookie.split("=", 1)[1].split(";", 1)[0]
@@ -1066,15 +1166,15 @@ class Ksphere:
         cookie_text = " " + cookie_value + "; $" + cookie_path
         headers = {'Content-Type': 'application/octet-stream', 'Cookie': f"{cookie_name}={cookie_text}"}
         with open(origin, "rb") as f:
-            ssl._create_default_https_context = ssl._create_unverified_context
+            context = ssl._create_unverified_context()
             try:
                 req = urllib.request.Request(url, data=f, headers=headers, method='PUT')
-                urllib.request.urlopen(req)
+                urllib.request.urlopen(req, context=context)
             except:
                 url = url.replace('/folder', '')
                 req = urllib.request.Request(url, data=f, headers=headers, method='PUT')
                 try:
-                    urllib.request.urlopen(req)
+                    urllib.request.urlopen(req, context=context)
                 except Exception as e:
                     error(f"Hit issue with with reason: {e}")
 
@@ -1087,7 +1187,9 @@ class Ksphere:
         vmFolder = self.basefolder
         vm, info = findvm2(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         config = info['config']
         spec = vim.vm.ConfigSpec()
         unit_number = 0
@@ -1138,6 +1240,43 @@ class Ksphere:
         error(msg)
         return {'result': 'failure', 'reason': error}
 
+    def detach_disks(self, name):
+        pprint(f"Detaching data disks from {name}")
+        si = self.si
+        dc = self.dc
+        vmFolder = self.basefolder
+        vm, info = findvm2(si, vmFolder, name)
+        if vm is None:
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
+        config = info['config']
+        disks = [dev for dev in config.hardware.device if isinstance(dev, vim.vm.device.VirtualDisk)]
+        if len(disks) > 1:
+            for index, disk in enumerate(disks[1:]):
+                devspec = vim.vm.device.VirtualDeviceSpec()
+                devspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
+                devspec.device = disk
+                spec = vim.vm.ConfigSpec()
+                spec.deviceChange = [devspec]
+                t = vm.ReconfigVM_Task(spec=spec)
+                waitForMe(t)
+                datastore = disk.backing.datastore.name
+                path = f"[{datastore}] {name}"
+                keep_path = f"[{datastore}] {name}_keep"
+                fileManager = self.si.content.fileManager
+                if index == 0:
+                    fileManager.MakeDirectory(name=keep_path, datacenter=dc)
+                file_name = f"{name}_{index + 1}"
+                for entry in [f"{file_name}.vmdk", f"{file_name}-flat.vmdk"]:
+                    t = fileManager.MoveDatastoreFile_Task(sourceName=f"{path}/{entry}",
+                                                           sourceDatacenter=dc,
+                                                           destinationName=f"{keep_path}/{entry}",
+                                                           destinationDatacenter=dc,
+                                                           force=True)
+                    WaitForTask(t)
+        return {'result': 'success'}
+
     def add_nic(self, name, network, model='virtio'):
         if network == 'default':
             network = 'VM Network'
@@ -1145,7 +1284,9 @@ class Ksphere:
         vmFolder = self.basefolder
         vm, info = findvm2(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         config = info['config']
         spec = vim.vm.ConfigSpec()
         nicnumber = len([dev for dev in config.hardware.device if "addressType" in dir(dev)])
@@ -1162,7 +1303,9 @@ class Ksphere:
         vmFolder = self.basefolder
         vm, info = findvm2(si, vmFolder, name)
         if vm is None:
-            return {'result': 'failure', 'reason': f"VM {name} not found"}
+            msg = f"VM {name} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         config = info['config']
         for dev in config.hardware.device:
             if isinstance(dev, vim.vm.device.VirtualEthernetCard) and dev.deviceInfo.label == interface:
@@ -1209,14 +1352,18 @@ class Ksphere:
         rootFolder = self.rootFolder
         net = find(si, rootFolder, vim.Network, name)
         if net is not None:
-            return {'result': 'failure', 'reason': f"Network {name} already there"}
+            msg = f"Network {name} already there"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         o = si.content.viewManager.CreateContainerView(rootFolder, [vim.DistributedVirtualSwitch], True)
         dvnetworks = o.view
         o.Destroy()
         for dvnetw in dvnetworks:
             for portg in dvnetw.portgroup:
                 if portg.name == name:
-                    return {'result': 'failure', 'reason': f"Network {name} already there"}
+                    msg = f"Network {name} already there"
+                    error(msg)
+                    return {'result': 'failure', 'reason': msg}
         if overrides.get('distributed', False):
             pnic_specs = []
             dvs_host_configs = []
@@ -1258,14 +1405,15 @@ class Ksphere:
                 net = find(si, rootFolder, vim.Network, name)
                 net.Destroy()
             except:
-                return {'result': 'failure', 'reason': f"Network {name} not found"}
+                msg = f"Network {name} not found"
+                error(msg)
+                return {'result': 'failure', 'reason': msg}
         return {'result': 'success'}
 
     def vm_ports(self, name):
         return ['default']
 
-    def add_image(self, url, pool, short=None, cmd=None, name=None, size=None, convert=False):
-        downloaded = False
+    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
         si = self.si
         rootFolder = self.rootFolder
         clu = find(si, rootFolder, vim.ComputeResource, self.clu)
@@ -1284,16 +1432,19 @@ class Ksphere:
             pprint(f"Template {shortimage} already there")
             return {'result': 'success'}
         if not find(si, rootFolder, vim.Datastore, pool):
-            return {'result': 'failure', 'reason': f"Pool {pool} not found"}
+            msg = f"Pool {pool} not found"
+            error(msg)
+            return {'result': 'failure', 'reason': msg}
         if os.path.exists(url):
             pprint(f"Using {url} as path")
         elif not os.path.exists(f'/tmp/{shortimage}'):
-            downloaded = True
             pprint(f"Downloading locally {shortimage}")
             downloadcmd = f"curl -kLo /tmp/{shortimage} -f '{url}'"
             code = os.system(downloadcmd)
             if code != 0:
-                return {'result': 'failure', 'reason': "Unable to download indicated image"}
+                msg = "Unable to download indicated image"
+                error(msg)
+                return {'result': 'failure', 'reason': msg}
         else:
             pprint(f"Using found /tmp/{shortimage}")
         if iso:
@@ -1314,26 +1465,31 @@ class Ksphere:
             return {'result': 'success'}
         vmdk_path = None
         ovf_path = None
+        basedir = os.path.dirname(os.path.abspath(url)) if os.path.exists(url) else '/tmp'
         if url.endswith('zip'):
-            with ZipFile(f"/tmp/{shortimage}") as zipf:
+            with ZipFile(f"{basedir}/{shortimage}") as zipf:
                 for _fil in zipf.namelist():
                     if _fil.endswith('vmdk'):
-                        vmdk_path = f'/tmp/{_fil}'
+                        vmdk_path = f'{basedir}/{_fil}'
                     elif _fil.endswith('ovf'):
-                        ovf_path = f'/tmp/{_fil}'
+                        ovf_path = f'{basedir}/{_fil}'
                 if vmdk_path is None or ovf_path is None:
-                    return {'result': 'failure', 'reason': "Incorrect ova file"}
-                zipf.extractall('/tmp')
+                    msg = "Incorrect ova file"
+                    error(msg)
+                    return {'result': 'failure', 'reason': msg}
+                zipf.extractall(basedir)
         elif url.endswith('ova'):
-            with tarfile.open(f"/tmp/{shortimage}") as tar:
+            with tarfile.open(f"{basedir}/{shortimage}") as tar:
                 for _fil in [x.name for x in tar.getmembers()]:
                     if _fil.endswith('vmdk'):
-                        vmdk_path = f'/tmp/{_fil}'
+                        vmdk_path = f'{basedir}/{_fil}'
                     elif _fil.endswith('ovf'):
-                        ovf_path = f'/tmp/{_fil}'
+                        ovf_path = f'{basedir}/{_fil}'
                 if vmdk_path is None or ovf_path is None:
-                    return {'result': 'failure', 'reason': "Incorrect ova file"}
-                tar.extractall('/tmp')
+                    msg = "Incorrect ova file"
+                    error(msg)
+                    return {'result': 'failure', 'reason': msg}
+                tar.extractall(basedir)
         else:
             need_uncompress = any(shortimage.endswith(suffix) for suffix in ['.gz', '.xz', '.bz2', '.zst'])
             if need_uncompress:
@@ -1341,16 +1497,24 @@ class Ksphere:
                 executable = {'xz': 'unxz', 'gz': 'gunzip', 'bz2': 'bunzip2', 'zst': 'zstd'}
                 flag = '--decompress' if extension == 'zstd' else '-f'
                 executable = executable[extension]
-                uncompresscmd = f"{executable} {flag} /tmp/{shortimage}"
+                uncompresscmd = f"{executable} {flag} {basedir}/{shortimage}"
                 os.system(uncompresscmd)
                 shortimage = shortimage.replace(f'.{extension}', '')
-            extension = os.path.splitext(shortimage)[1].replace('.', '')
-            vmdk_file = shortimage.replace(extension, 'vmdk')
-            vmdk_path = f"/tmp/{vmdk_file}"
+            if '.' in shortimage:
+                extension = os.path.splitext(shortimage)[1].replace('.', '')
+                vmdk_file = shortimage.replace(extension, 'vmdk')
+            else:
+                vmdk_file = f"{shortimage}.vmdk"
+            vmdk_path = f"{basedir}/{vmdk_file}"
+            if cmds and shortimage.endswith('qcow2') and which('virt-customize') is not None:
+                for cmd in cmds:
+                    cmd = f"virt-customize -a {basedir}/{shortimage} --run-command '{cmd}'"
+                    os.system(cmd)
             if not os.path.exists(vmdk_path):
                 pprint("Converting qcow2 file to vmdk")
-                os.popen(f"qemu-img convert -O vmdk -o subformat=streamOptimized /tmp/{shortimage} {vmdk_path}").read()
-            ovf_path = "/tmp/%s" % shortimage.replace(extension, 'ovf')
+                cmd = f"qemu-img convert -O vmdk -o subformat=streamOptimized {basedir}/{shortimage} {vmdk_path}"
+                os.popen(cmd).read()
+            ovf_path = vmdk_path.replace('.vmdk', '.ovf')
             commondir = os.path.dirname(common.pprint.__code__.co_filename)
             vmdk_info = json.loads(os.popen(f"qemu-img info {vmdk_path} --output json").read())
             virtual_size = vmdk_info['virtual-size']
@@ -1371,7 +1535,7 @@ class Ksphere:
                 network = networks[0]
                 break
         if network is None:
-            error("Couldn't find any esxi host in the cluster with network {self.import_network}")
+            error(f"Couldn't find any esxi host in the cluster with network {self.import_network}")
             error("Set import_network to a valid value")
             sys.exit(1)
         networkmapping = vim.OvfManager.NetworkMapping.Array()
@@ -1386,23 +1550,24 @@ class Ksphere:
         lease = resourcepool.ImportVApp(import_spec.importSpec, vmFolder)
         while True:
             if lease.state == vim.HttpNfcLease.State.ready:
-                pprint("Uploading vmdk")
                 url = lease.info.deviceUrl[0].url.replace('*', host.name)
+                pprint(f"Uploading {vmdk_path} to {url}")
                 keepalive_thread = Thread(target=keep_lease_alive, args=(lease,))
                 keepalive_thread.start()
-                upload_cmd = (
-                    "curl -sS -X POST --insecure -T %s -H 'Content-Type: \
-                    application/x-vnd.vmware-streamVmdk' %s" % (vmdk_path, url))
-                os.system(upload_cmd)
+                cmd = f'curl -k -# -X POST -T {vmdk_path} -H "Content-Type: application/x-vnd.vmware-streamVmdk" {url}'
+                call(cmd, shell=True)
                 lease.HttpNfcLeaseComplete()
                 keepalive_thread.join()
                 break
             elif lease.state == vim.HttpNfcLease.State.error:
                 error(f"Lease error: {lease.error}")
                 sys.exit(1)
-        self.convert_to_template(name)
-        if downloaded:
-            os.remove(f'/tmp/{shortimage}')
+        if not self.esx:
+            self.convert_to_template(name)
+            if os.path.exists(f'{basedir}/{shortimage}'):
+                os.remove(f'{basedir}/{shortimage}')
+            if os.path.exists(vmdk_path):
+                os.remove(vmdk_path)
         return {'result': 'success'}
 
     def _get_hosts(self, cluster):
@@ -1454,9 +1619,17 @@ class Ksphere:
         return data
 
     def delete_image(self, image, pool=None):
+        si = self.si
+        if image.endswith('.iso'):
+            matching_isos = [iso for iso in self._getisos() if iso.endswith(image)]
+            if not matching_isos:
+                return {'result': 'failure', 'reason': f'Iso {image} not found'}
+            else:
+                isopath = matching_isos[0]
+                deletedirectory(si, self.dc, isopath)
+                return {'result': 'success'}
         if '/' in image:
             pool, image = image.split('/')
-        si = self.si
         vmFolder = self.basefolder
         vm, info = findvm2(si, vmFolder, image)
         if vm is None or not info['config'].template:
@@ -1518,7 +1691,7 @@ class Ksphere:
         elif 'debian' in image.lower():
             guestid = 'debian10_64Guest'
         else:
-            guestid = 'rhel7_64Guest'
+            guestid = 'genericLinuxGuest'
         return guestid
 
     def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False):
@@ -1687,6 +1860,8 @@ class Ksphere:
         datastores = {}
         for datacenter in view:
             for clu in datacenter.hostFolder.childEntity:
+                if 'ClusterComputeResource' not in str(clu.__class__):
+                    continue
                 for datastore in clu.datastore:
                     datastores[datastore.name] = datacenter.name
         return datastores
@@ -1710,3 +1885,27 @@ class Ksphere:
     def list_dns_zones(self):
         print("not implemented")
         return []
+
+    def set_macs(self, name, macs):
+        vm = find(self.si, self.dc.vmFolder, vim.VirtualMachine, name)
+        currentdevices = vm.config.hardware.device
+        currentnics = [d for d in currentdevices if isinstance(d, vim.vm.device.VirtualEthernetCard)]
+        confspec = vim.vm.ConfigSpec()
+        devconfspec = []
+        for index in macs:
+            currentnic = currentnics[index]
+            currentnic.macAddress = macs[index]
+            currentnic.addressType = vim.vm.device.VirtualEthernetCardOption.MacTypes.manual
+            nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
+            devconfspec.append(nicspec)
+        confspec.deviceChange = devconfspec
+        t = vm.Reconfigure(confspec)
+        waitForMe(t)
+
+    def reconnect_hosts(self):
+        si = self.si
+        rootFolder = self.rootFolder
+        view = si.content.viewManager.CreateContainerView(rootFolder, [vim.HostSystem], True)
+        for host in view.view:
+            pprint(f"Reconnecting Host {host.name}")
+            host.Reconnect()
